
;---------------------------------------------------------------------------
;     FILL THE MISSING PIXELS IN ONE IMAGE USING A TIME SERIES AS INPUT
;                                        
;                     Can be used for entire ETM+ scene
;            Developed by Zhu Xiaolin,email: zhuxiaolin55@gmail.com
;                   Hong Kong Polytechnic Unviersity
;                            11/20/2017
;    
;                     Copyright belongs to Xiaolin Zhu
;                     
;     Updated: 05/06/2015: modified the original NSPI code to fill gaps, clouds,
;                          shadows in all ETM+ images in a time series
;              07/11/2017: debug when no similar pixel selected
;              03/10/2018: improve efficiency by reducing loops
;              03/14/2018: speedup the similar pixel searching by binary search for large cloud patches
;     
;                     Copyright belongs to Xiaolin Zhu
;
;Please cite:
;1)Chen, J., Zhu, X. Vogelmann, J. E. Gao, F. and Jin, S. A simple and effective 
;method for filling gaps in Landsat ETM+ SLC-off images. 
;Remote Sensing of  Environment, 2011, 115, 1053-1064 
;2)Zhu, X., Gao, F., Liu, D. and Chen, J. A modified neighborhood similar pixel 
;interpolator approach for removing thick clouds in Landsat images. 
;IEEE Geoscience  and Remote Sensing Letters, 2012, 9(3), 521-525
;3)Zhu, X., Helmer, E.H., Chen, J., and Liu, D. An automatic system for reconstructing 
;High-quality seasonal Landsat time series, in Remote Sensing Time Series Image Processing,
;CRC Press,2018                   
;---------------------------------------------------------------------------


;function for open the file

Pro GetData,ImgData = ImgData,ns = ns,nl = nl,nb = nb,Data_Type = Data_Type,$
    FileName = FileName,Map_info = map_Info, Fid = Fid
    Filter = ['all file;*.*']
    Envi_Open_File,FileName,R_Fid = Fid
    Envi_File_Query,Fid,ns = ns,nl = nl,nb = nb,Data_Type = Data_Type
    map_info = envi_get_map_info(fid=Fid)
    dims = [-1,0,ns - 1 ,0,nl - 1]
    case Data_Type Of
        1:ImgData = BytArr(ns,nl,nb)    ;  BYTE  Byte
        2:ImgData = IntArr(ns,nl,nb)    ;  INT  Integer
        3:ImgData = LonArr(ns,nl,nb)    ;  LONG  Longword integer
        4:ImgData = FltArr(ns,nl,nb)    ;  FLOAT  Floating point
        5:ImgData = DblArr(ns,nl,nb)    ;  DOUBLE  Double-precision floating
        6:ImgData = COMPLEXARR(ns,nl,nb); complex, single-precision, floating-point
        9:ImgData = DCOMPLEXARR(ns,nl,nb);complex, double-precision, floating-point
        12:ImgData = UINTARR(ns,nl,nb)   ; unsigned integer vector or array
        13:ImgData = ULONARR(ns,nl,nb)   ;  unsigned longword integer vector or array
        14:ImgData = LON64ARR(ns,nl,nb)   ;a 64-bit integer vector or array
        15:ImgData = ULON64ARR(ns,nl,nb)   ;an unsigned 64-bit integer vector or array
    EndCase
    For i = 0,nb-1 Do Begin
       Dt = Envi_Get_Data(Fid = Fid,dims = dims,pos=i)
       ImgData[*,*,i] = Dt[*,*]
    EndFor
End

;-------------------------------------------------------------------
;                  main body of the program
;-------------------------------------------------------------------

pro  NSPI_fillSINGLE_useTIMESERIES

 t0=systime(1)                  ;the initial time of program running

 ;please set the following parameters
;----------------------------------------------------------------------
 min_similar=30                 ;set the minimum sample size of similar pixels, 20~40 recommended
 num_class=4                    ;set the number of classes (given empirically)
 num_band=6                     ;set the number of bands in each images
 DN_min=0.0                     ;set the range of DN value of the image,If byte, set 0 and 255
 DN_max=10000.0
 patch_long=500                 ;set the size of block,if process whole Landsat scene, 500-1000 recommended
 temp_file='D:\temp'            ;set the address of the temporary folder
 Target=5                       ;set the target image in the time series to be reconstructed
 DOY=[17,33,40,56,65,97,104,136,145,168,184,193,200,209,216]  ;Day of year of each image in the time-series
;------------------------------------------------------------------------


 ;open the SLC-off ETM+ time-series image
  FileName1 = Dialog_PickFile(title = 'Open the original time-series images:')
  envi_open_file,FileName1,r_fid=fid
  envi_file_query,fid,ns=ns,nl=nl,nb=nb,dims=dims
  map_info = envi_get_map_info(fid=fid)
  orig_ns=ns
  orig_nl=nl
  n_ns=ceil(float(ns)/patch_long)
  n_nl=ceil(float(nl)/patch_long)
  
  n_image=ceil(nb/num_band)

  ind_patch1=intarr(4,n_ns*n_nl)           ;divide the whole scene into 1000*1000 block
  ind_patch=intarr(4,n_ns*n_nl)
  location=intarr(4,n_ns*n_nl)

  for i_ns=0,n_ns-1,1 do begin
    for i_nl=0,n_nl-1,1 do begin
        ind_patch1[0,n_ns*i_nl+i_ns]=i_ns*patch_long
        ind_patch[0,n_ns*i_nl+i_ns]=max([0,ind_patch1[0,n_ns*i_nl+i_ns]-10])
        location[0,n_ns*i_nl+i_ns]=ind_patch1[0,n_ns*i_nl+i_ns]-ind_patch[0,n_ns*i_nl+i_ns]

        ind_patch1[1,n_ns*i_nl+i_ns]=min([ns-1,(i_ns+1)*patch_long-1])
        ind_patch[1,n_ns*i_nl+i_ns]=min([ns-1,ind_patch1[1,n_ns*i_nl+i_ns]+10])
        location[1,n_ns*i_nl+i_ns]=ind_patch1[1,n_ns*i_nl+i_ns]-ind_patch1[0,n_ns*i_nl+i_ns]+location[0,n_ns*i_nl+i_ns]

        ind_patch1[2,n_ns*i_nl+i_ns]=i_nl*patch_long
        ind_patch[2,n_ns*i_nl+i_ns]=max([0,ind_patch1[2,n_ns*i_nl+i_ns]-10])
        location[2,n_ns*i_nl+i_ns]=ind_patch1[2,n_ns*i_nl+i_ns]-ind_patch[2,n_ns*i_nl+i_ns]

        ind_patch1[3,n_ns*i_nl+i_ns]=min([nl-1,(i_nl+1)*patch_long-1])
        ind_patch[3,n_ns*i_nl+i_ns]=min([nl-1,ind_patch1[3,n_ns*i_nl+i_ns]+10])
        location[3,n_ns*i_nl+i_ns]=ind_patch1[3,n_ns*i_nl+i_ns]-ind_patch1[2,n_ns*i_nl+i_ns]+location[2,n_ns*i_nl+i_ns]
    endfor
  endfor

  tempoutname=temp_file+'\temp_target'
 
  pos=indgen(nb)
  for isub=0,n_ns*n_nl-1,1 do begin
      dims=[-1,ind_patch[0,isub],ind_patch[1,isub],ind_patch[2,isub],ind_patch[3,isub]]
      envi_doit, 'resize_doit', fid=fid, pos=pos, dims=dims, interp=0, rfact=[1,1], $
      out_name=tempoutname+strtrim(isub+1,1), r_fid=r_fid1
      envi_file_mng, id=r_fid1, /remove
  endfor

  envi_file_mng, id=fid, /remove

  ;-----------------------------------------------------------     
  ;open the mask of SLC-off ETM+ image: 0 is good pixels, 1 gaps 2 shadows 3 clouds

   FileName2 = Dialog_PickFile(title = 'Open the corresponding mask images:')
   envi_open_file,FileName2,r_fid=fid
   tempoutname=temp_file+'\temp_mask'
   pos=indgen(n_image)
   for isub=0,n_ns*n_nl-1,1 do begin
      dims=[-1,ind_patch[0,isub],ind_patch[1,isub],ind_patch[2,isub],ind_patch[3,isub]]
      envi_doit, 'resize_doit', fid=fid, pos=pos, dims=dims, interp=0, rfact=[1,1], $
      out_name=tempoutname+strtrim(isub+1,1), r_fid=r_fid1
      envi_file_mng, id=r_fid1, /remove
  endfor
   envi_file_mng, id=fid, /remove


;------------------------------------------------------------------
        ; begin process the gap for each block
;-------------------------------------------------------------------

print,'there are total',n_ns*n_nl,' blocks'

for isub=0,n_ns*n_nl-1,1 do begin

;open each block image

    FileName = temp_file+'\temp_target'
    GetData,ImgData=ImgData,ns = ns,nl = nl,nb = nb,Data_Type = Data_Type,FileName = FileName+strtrim(isub+1,1),Fid = Fid1
    fine1_all=float(ImgData)
    fine0=fine1_all[location[0,isub]:location[1,isub],location[2,isub]:location[3,isub],(Target-1)*num_band:(Target*num_band-1)]    ;place the new image value

    FileName = temp_file+'\temp_mask'
    GetData,ImgData=ImgData,FileName = FileName+strtrim(isub+1,1),Fid = Fid2
    mask=float(ImgData)

    mark=intarr(location[1,isub]-location[0,isub]+1,location[3,isub]-location[2,isub]+1,1)     ;place the mark value
    
; classify all images by k-means
class=intarr(ns,nl,n_image)
for i_img=0, n_image-1, 1 do begin
  imagei=fine1_all[*,*,i_img*num_band:(i_img*num_band+num_band-1)]
  gapi=mask[*,*,i_img]
  ind_bad=where(gapi ne 0, num_bad)
  if (num_bad gt 0) then begin
    for iband=0,num_band-1,1 do begin
      temp=imagei[*,*,iband]
      temp[ind_bad]=0.0
      imagei[*,*,iband]=temp
    endfor
  endif
  
  tempnamei=temp_file+'\imagei'
  Envi_Write_Envi_File,imagei,Out_Name = tempnamei
  envi_open_file,tempnamei,r_fid=fid5
  
  ;parameter of isodata
  CHANGE_THRESH = .02
  NUM_CLASSES = num_class+1
  ITERATIONS = 20
  out_bname = 'kmeans'
  out_name=temp_file+'\class_kmeans'
  ENVI_File_Query, fid5, DIMS=dims, NB=nb1
      ENVI_DOIT, 'class_doit', fid=fid5, pos=indgen(nb1), dims=dims, $
        out_bname=out_bname, out_name=out_name, method=7, $
        r_fid=r_fid, $
        lookup = BYTARR(3,num_classes+1), $
        NUM_CLASSES = NUM_CLASSES, $
        in_memory=0, CHANGE_THRESH=CHANGE_THRESH,$
        ITERATIONS = ITERATIONS
    
    GetData,ImgData=class_img,FileName = out_name,Fid = Fid6
    
    class[*,*,i_img]=reform(class_img,ns,nl)
    envi_file_mng, id=fid5, /remove,/delete
    envi_file_mng, id=fid6, /remove,/delete
endfor

;FileName ='C:\Users\lsxlzhu.LSGI\Desktop\temp\class'
;Envi_Write_Envi_File,class,Out_Name = FileName,Map_info = map_Info
    
;process the target image
for i_img=Target-1, Target-1, 1 do begin
  
    fine1=fine1_all[*,*,i_img*num_band:(i_img*num_band+num_band-1)]
    gap=mask[*,*,i_img]
    temp_dis=abs(DOY-DOY[i_img])
    order=sort(temp_dis)
  
    note_finish=0
    i_input=1

 while (note_finish ne 1 and i_input le n_image-1) do begin     
       
       fine2=fine1_all[*,*,order[i_input]*num_band:(order[i_input]*num_band+num_band-1)]
       mask2=mask[*,*,order[i_input]]
       class2=class[*,*,order[i_input]]
      
       similar_th_band=fltarr(num_band)
       ind_nogrand=where(mask2 eq 0,c_nogrand)
       if (c_nogrand gt 0) then begin
        for iband=0,num_band-1,1 do begin
            similar_th_band[iband]=stddev((fine2[*,*,iband])[ind_nogrand])*2.0/float(num_class)    ;compute the threshold of similar pixel          
        endfor
       endif
       similar_th=mean(similar_th_band)
          

for i=location[0,isub],location[1,isub],1 do begin
  for j=location[2,isub],location[3,isub],1 do begin

    if (gap[i,j] gt 0 and gap[i,j] lt 10) then begin  ;only process contaminated pixels exclude background pixels
        
      
       if (mask2[i,j] eq 0) then begin               ;if the corresponding pixel avalible,use this input image
          
           classij=class2[i,j]          
           extent=ceil(0.5*(min_similar^0.5-1)) ;compute the minimum window size 
                       
          max_window= round(max([ns,nl])*0.25) ; maximum size of half seach window is 1/4 of image size
        
        
        ;find the possible largest window
        gap_profiles=[gap[*,j],reform(gap[i,*])]                ;check vertical and horizontal directions
        input_profile=[mask2[*,j],reform(mask2[i,*])]
        spatial_position=abs([indgen(ns)-i,indgen(nl)-j])     
        ind_outcloud=where(gap_profiles eq 0 and input_profile eq 0, num_outcloud)
        if (num_outcloud gt 0) then begin
          max_window_p=min(spatial_position[ind_outcloud])    ;find the possible maximum half window
          max_window_p=min([max_window,max_window_p+50])
        endif else begin
          max_window_p=max_window          
        endelse
      
        start_w=ceil(0.5*(min_similar^0.5-1))
        ;check large cloud patches to avoid searching similar pixels for it
        if (max_window_p eq max_window) then begin
                  a1=max([0,i-max_window])
                  a2=min([ns-1,i+max_window])
                  b1=max([0,j-max_window])
                  b2=min([nl-1,j+max_window])
                  sub_gap=gap[a1:a2,b1:b2]
                  sub_mask2=mask2[a1:a2,b1:b2]
                  sub_class=class2[a1:a2,b1:b2]
                  ind_common=where(sub_gap eq 0 and sub_mask2 eq 0 and sub_class eq classij, c_common)
                  if (c_common lt 2.0*min_similar) then begin
                        start_w=max_window_p
                  endif
        endif
            
        ; use binary search to find the best window size      
        end_w=max_window_p
        ind_sucess=0
        while (ind_sucess eq 0 and start_w le end_w)do begin 
          mid=floor(start_w+(end_w-start_w)/2.0)         
          a1=max([0,i-mid])
          a2=min([ns-1,i+mid])
          b1=max([0,j-mid])
          b2=min([nl-1,j+mid])
          sub_gap=gap[a1:a2,b1:b2]
          sub_mask2=mask2[a1:a2,b1:b2]
          sub_class=class2[a1:a2,b1:b2]
          ind_common=where(sub_gap eq 0 and sub_mask2 eq 0 and sub_class eq classij, c_common)
          if (c_common gt 3.0*min_similar) then begin ; search left range
            end_w=mid-1
          endif else begin
            if (c_common lt 2.0*min_similar) then begin ;searcg right range
               start_w=mid+1
            endif else begin
               ind_sucess=1
            endelse
          endelse 
        endwhile
        
           extent=mid

           sub_off=fine1[a1:a2,b1:b2,*]
           sub_on=fine2[a1:a2,b1:b2,*]
           
           col_wind=indgen(a2-a1+1)#(intarr(1,b2-b1+1)+1)*1.0
           row_wind=(intarr(a2-a1+1)+1)#indgen(1,b2-b1+1)*1.0              
           it=i-a1
           jt=j-b1

           if (c_common ge 2.0*min_similar) then begin  ;if same-class pixels more than required sample size

                 disi=((it[0]-col_wind[ind_common])^2+(jt[0]-row_wind[ind_common])^2)^0.5
                 sub_on_common=fltarr(c_common,num_band)
                 sub_off_common=fltarr(c_common,num_band)
                 for iband=0,num_band-1,1 do begin
                     sub_on_common[*,iband]= (sub_on[*,*,iband])[ind_common]-fine2[i,j,iband]
                     sub_off_common[*,iband]= (sub_on[*,*,iband])[ind_common]-(sub_off[*,*,iband])[ind_common]
                 endfor
                 rmsei=mean(sub_on_common^2,DIMENSION=2)^0.5+0.0001
                 rmse12=mean(sub_off_common^2,DIMENSION=2)^0.5+0.0001
                
                  order_rmse=sort(rmsei)
                  ind_similar=order_rmse[0:(min_similar-1)]    ;find similar pixels
             
                  similar_rmse=rmsei[ind_similar]
                  similar_rmse12=rmse12[ind_similar]                                         ;compute the weight W
                  similar_dis=disi[ind_similar]
                  C_D=similar_rmse*similar_dis
                  weight=(1.0/C_D)/total(1.0/C_D)
                  T_1=mean(similar_rmse)                                     ;compute the weight T
                  T_2=mean(similar_rmse12)
                  W_T1=T_2/(T_1+T_2)
                  W_T2=T_1/(T_1+T_2)
                  for iband=0,num_band-1,1 do begin
                      similar_on=((sub_on[*,*,iband])[ind_common])[ind_similar]
                      similar_off=((sub_off[*,*,iband])[ind_common])[ind_similar]
                      predict_1=total(similar_off*weight)
                      predict_2=fine2[i,j,iband]+total((similar_off-similar_on)*weight)
                       if (predict_2 gt DN_min and predict_2 lt DN_max) then begin
                           fine0[i-location[0,isub],j-location[2,isub],iband]=W_T1*predict_1+W_T2*predict_2               
                       endif else begin
                           fine0[i-location[0,isub],j-location[2,isub],iband]=predict_1
                       endelse
                   endfor
                   note=1
                   mark[i-location[0,isub],j-location[2,isub],0]=1+10*i_input
                   gap[i,j]=-2                            
                endif else begin                 ;if no enough same-class pixels exist
                       if (c_common ge 3 and extent ge max_window) then begin   ;IF reach the maximum window, use all the same-class pixels                           
                         disi=((it[0]-col_wind[ind_common])^2+(jt[0]-row_wind[ind_common])^2)^0.5
                      
                         sub_on_common=fltarr(c_common,num_band)
                         sub_off_common=fltarr(c_common,num_band)
                          for iband=0,num_band-1,1 do begin
                           sub_on_common[*,iband]= (sub_on[*,*,iband])[ind_common]-fine2[i,j,iband]
                           sub_off_common[*,iband]= (sub_on[*,*,iband])[ind_common]-(sub_off[*,*,iband])[ind_common]
                          endfor
                          rmsei=mean(sub_on_common^2,DIMENSION=2)^0.5+0.0001
                          rmse12=mean(sub_off_common^2,DIMENSION=2)^0.5+0.0001
                         
                          similar_rmse=rmsei
                          similar_rmse12=rmse12
                          similar_dis=disi
                          
                          C_D=similar_rmse*similar_dis
                          weight=(1.0/C_D)/total(1.0/C_D)
                        ;  print,'weight',weight
                          
                          T_1=mean(similar_rmse)
                          T_2=mean(similar_rmse12)
                          W_T1=T_2/(T_1+T_2)
                          W_T2=T_1/(T_1+T_2)
                          for iband=0,num_band-1,1 do begin
                             similar_on=(sub_on[*,*,iband])[ind_common]
                             similar_off=(sub_off[*,*,iband])[ind_common]
                             predict_1=total(similar_off*weight)
                             predict_2=fine2[i,j,iband]+total((similar_off-similar_on)*weight)
                     
                             if (predict_2 gt DN_min and predict_2 lt DN_max) then begin
                                fine0[i-location[0,isub],j-location[2,isub],iband]=W_T1*predict_1+W_T2*predict_2
                             endif else begin
                                fine0[i-location[0,isub],j-location[2,isub],iband]=predict_1
                             endelse
                           endfor
                          mark[i-location[0,isub],j-location[2,isub],0]=2+10*i_input
                          note=1
                          gap[i,j]=-2                    
                        endif else begin  ; if no same-class pixels exist when reaching the maximum window, directly replace the pixel by input image  
                              ;only use temporal interpolation if it have observations before and after
                              first_half=max([0,i_img-1])
                              second_half=min([i_img+1,n_image-1])
                              ind_good_before=where(mask[i,j,0:first_half] eq 0, num_good_before)
                              ind_good_after= where(mask[i,j,second_half:n_image-1] eq 0, num_good_after) 
                              if (num_good_before gt 0 and num_good_after gt 0) then begin 
                                 i_before=ind_good_before[num_good_before-1]
                                 i_after= ind_good_after[0]+i_img+1                                                         
                                  T1=doy[i_img]-doy[i_before]
                                  T2=doy[i_after]-doy[i_img]       
                                 for iband=0,num_band-1,1 do begin
                                     before_band=fine1_all[i,j,i_before*num_band+iband] 
                                     after_band= fine1_all[i,j,i_after*num_band+iband]                                            
                                     fine0[i-location[0,isub],j-location[2,isub],iband]=(T2*before_band+T1*after_band)/(T2+T1)                                            
                                 endfor
                              endif else begin
                                  fine0[i-location[0,isub],j-location[2,isub],*]=fine2[i,j,*] 
                              endelse                         
                             mark[i-location[0,isub],j-location[2,isub],0]=3+10*i_input
                             note=1
                             gap[i,j]=-2  
                        endelse  
                                            
                    endelse              
             
        endif       
       endif
     endfor
    endfor

      ind_unfilled=where(gap[location[0,isub]:location[1,isub],location[2,isub]:location[3,isub]]  gt 0, c_unfilled)
      if (c_unfilled eq 0) then begin
         note_finish=1
      endif else begin
         i_input=i_input+1
      endelse
       
  endwhile
   
   
   ;print,'finish ',i_img+1,' image in the time-series'
   
   
 endfor

   size_result=size(fine0)

    case Data_Type Of
        1:fine0 = Byte(fine0)    ;  BYTE  Byte
        2:fine0 = FIX(fine0)     ;  INT  Integer
        3:fine0 = LONG(fine0)    ;  LONG  Longword integer
        4:fine0 = FLOAT(fine0)   ;  FLOAT  Floating point
        5:fine0 = DOUBLE(fine0)  ;  DOUBLE  Double-precision floating
        6:fine0 = COMPLEX(fine0); complex, single-precision, floating-point
        9:fine0 = DCOMPLEX(fine0);complex, double-precision, floating-point
        12:fine0 = UINT(fine0)   ; unsigned integer vector or array
        13:fine0 = ULONG(fine0)   ;  unsigned longword integer vector or array
        14:fine0 = LONG64(fine0)   ;a 64-bit integer vector or array
        15:Ifine0 = ULONG64(fine0)   ;an unsigned 64-bit integer vector or array
    EndCase

       print,'finished ',isub+1,' block'
       
         tempoutname1=temp_file+'\temp_filled'
         tempoutname2=temp_file+'\temp_mark'       
         Envi_Write_Envi_File,fine0,Out_Name = tempoutname1+strtrim(isub+1,1)
         Envi_Write_Envi_File,mark,Out_Name = tempoutname2+strtrim(isub+1,1)
         envi_file_mng, id=Fid1, /remove, /delete
         envi_file_mng, id=Fid2, /remove, /delete

endfor

;--------------------------------------------------------------------------------------
;;mosiac all the filled patch

  mfid=intarr(n_ns*n_nl)
  mdims=intarr(5,n_ns*n_nl)
  mpos=intarr(num_band,n_ns*n_nl)
  pos=indgen(num_band)
  x0=intarr(n_ns*n_nl)
  y0=intarr(n_ns*n_nl)

  for isub=0,n_ns*n_nl-1,1 do begin
      envi_open_file, tempoutname1+strtrim(isub+1,1), r_fid= sub_fid
     if (sub_fid eq -1) then begin
       envi_batch_exit
       return
     endif
      envi_file_query,  sub_fid, ns=sub_ns, nl=sub_nl
      mfid[isub] = sub_fid
      mpos[*,isub] = indgen(num_band)
      mdims[*,isub] = [-1,0, sub_ns-1,0, sub_nl-1]
      x0[isub]=ind_patch1[0,isub]
      y0[isub]=ind_patch1[2,isub]
  endfor

  xsize = orig_ns
  ysize = orig_nl
  pixel_size = [1.,1.]

  use_see_through = replicate(1L,n_ns*n_nl)
  see_through_val = replicate(0L,n_ns*n_nl)

;   out_name=Dialog_PickFile(Title = 'Enter the filename of the gap filled image')
    out_name=FileName1+'_image'+strtrim(Target,1)+'_filled_NSPI'
    envi_doit, 'mosaic_doit', fid=mfid, pos=mpos, $
    dims=mdims, out_name=out_name, xsize=xsize, $
    ysize=ysize, x0=x0, y0=y0, georef=0,MAP_INFO=map_info, $
    out_dt=Data_Type, pixel_size=pixel_size, $
    background=0, see_through_val=see_through_val, $
    use_see_through=use_see_through

    for i=0,n_ns*n_nl-1,1 do begin
      envi_file_mng, id=mfid[i], /remove, /delete
    endfor

;--------------------------------------------------------------------------------------
;mosiac all the mask patch

  mfid=intarr(n_ns*n_nl)
  mdims=intarr(5,n_ns*n_nl)
  mpos=intarr(1,n_ns*n_nl)
  pos=indgen(1)
  x0=intarr(n_ns*n_nl)
  y0=intarr(n_ns*n_nl)

  for isub=0,n_ns*n_nl-1,1 do begin
      envi_open_file, tempoutname2+strtrim(isub+1,1), r_fid= sub_fid
     if (sub_fid eq -1) then begin
       envi_batch_exit
       return
     endif
      envi_file_query,  sub_fid, ns=sub_ns, nl=sub_nl
      mfid[isub] = sub_fid
      mpos[*,isub] = indgen(1)
      mdims[*,isub] = [-1,0, sub_ns-1,0, sub_nl-1]
      x0[isub]=ind_patch1[0,isub]
      y0[isub]=ind_patch1[2,isub]
  endfor

  xsize = orig_ns
  ysize = orig_nl
  pixel_size = [1.,1.]

  use_see_through = replicate(1L,n_ns*n_nl)
  see_through_val = replicate(0L,n_ns*n_nl)

;   out_name=Dialog_PickFile(Title = 'Enter the filename of the mask image')
    out_name=FileName1+'_image'+strtrim(Target,1)+'_fill_quality_flag'
    envi_doit, 'mosaic_doit', fid=mfid, pos=mpos, $
    dims=mdims, out_name=out_name, xsize=xsize, $
    ysize=ysize, x0=x0, y0=y0, georef=0,MAP_INFO=map_info, $
    out_dt=1, pixel_size=pixel_size, $
    background=0, see_through_val=see_through_val, $
    use_see_through=use_see_through

  for i=0,n_ns*n_nl-1,1 do begin
    envi_file_mng, id=mfid[i], /remove, /delete
  endfor

print, 'time used:', floor((systime(1)-t0)/3600), 'h',floor(((systime(1)-t0) mod 3600)/60),'m',(systime(1)-t0) mod 60,'s'


end